/*-
 * Copyright (c) 1999-2002 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
  "@(#)$Id: kipfw.c,v 1.4.2.5 2002/09/24 10:52:30 simon Exp $";
#endif /* !lint */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <varargs.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "kipfw.h"

#include "debug.h"


u_quad_t	kipfw_bcnt_max;
struct ip_fw	*kipfw;
u_int		nkipfw;

static int	sd = -1;
static int	nbytes;
static u_int	nalloc;

static char	ipfw_errmsg[] = "IPv4 Firewall is not configured in the kernel or does not work properly";


/*
 * Init IP Firewall support: open raw socket.
 */
int
kipfw_init(void)
{
	struct ip_fw	tmp;

	syslog(LOG_INFO, "preinit IPv4 Firewall support");
	if ( (sd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
		syslog(LOG_ERR, "kipfw_init: socket(AF_INET, SOCK_RAW, IPPROTO_RAW): %m");
		return -1;
	}
	kipfw = NULL;
	nkipfw = nalloc = 0;
	
	switch (sizeof(tmp.fw_bcnt)) {
	case 4:
		kipfw_bcnt_max = ULONG_MAX;
		break;
	case 8:
		kipfw_bcnt_max = UQUAD_MAX;
		break;
	default:
		syslog(LOG_ERR, "kipfw_init: unsupported size of fw_bcnt (%d bytes) in struct ip_fw",
		    sizeof tmp.fw_bcnt);
		return -1;
	}

	if (debug_ipfw > 0) {
		syslog(LOG_INFO, "kipfw_init: raw socket opened, FD %d", sd);
		if (debug_ipfw > 1)
			syslog(LOG_INFO, "kipfw_init: sizeof(struct ip_fw) = %u, sizeof(fw_bcnt) = %u",
			    sizeof tmp, sizeof tmp.fw_bcnt);
	}

	return 0;
}

/*
 * Read the table of IPFW rules from the kernel.
 * Reallocate internal structures if needed.
 */
int
kipfw_read_table(void)
{
	static int	nbytesalloc;

	if (kipfw != NULL) {
		nbytes = nbytesalloc;
		if (getsockopt(sd, IPPROTO_IP, IP_FW_GET, kipfw, &nbytes) < 0) {
			syslog(LOG_ERR, "kipfw_read_table: getsockopt(IP_FW_GET): %m");
			syslog(LOG_ERR, "%s", ipfw_errmsg);
			return -1;
		}
		nkipfw = nbytes / sizeof(struct ip_fw);
		if (debug_ipfw > 0)
			syslog(LOG_INFO, "kipfw_read_table: nkipfw = %u, nbytes = %u, nbytesalloc = %u",
			    nkipfw, nbytes, nbytesalloc);
	}
	while (nkipfw >= nalloc) {
		nalloc = nalloc * 2 + 200;
		nbytes = nbytesalloc = nalloc * sizeof(struct ip_fw);
		if ( (kipfw = realloc(kipfw, nbytes)) == NULL) {
			syslog(LOG_ERR, "kipfw_read_table: realloc(%d bytes): %m", nbytes);
			return -1;
		}
		if (getsockopt(sd, IPPROTO_IP, IP_FW_GET, kipfw, &nbytes) < 0) {
			syslog(LOG_ERR, "kipfw_read_table: getsockopt(IP_FW_GET): %m");
			syslog(LOG_ERR, "%s", ipfw_errmsg);
			return -1;
		}
		nkipfw = nbytes / sizeof(struct ip_fw);
		if (debug_ipfw > 0) {
			syslog(LOG_INFO, "kipfw_read_table: resize table");
			syslog(LOG_INFO, "kipfw_read_table: nkipfw = %u, nbytes = %u, nbytesalloc = %u, nalloc = %u",
			    nkipfw, nbytes, nbytesalloc, nbytesalloc / sizeof(struct ip_fw));
		}
	}
	if (debug_ipfw > 1) {
		u_int	i;
		char	one[10 + 5 + 1 + 1 + 1];
		/* 10 - length of u_int, 5 - length of u_short,
		   1 - '.', 1 - ' ', 1 - '\0' */
		int	len, total_len;
		u_int	subnumber = 0;
#define TXT_LEN 60
		char	txt[TXT_LEN + 1];
		int	prevnum = -1;
		struct ip_fw	*kipfwp = kipfw;

		syslog(LOG_INFO, "kipfw_read_table: dump IPFW table (%u record%s):",
		    nkipfw, nkipfw > 1 ? "s" : "");
		for (i = 0, total_len = 0; i < nkipfw; ++kipfwp, ++i) {
			if (prevnum == kipfwp->fw_number)
				++subnumber;
			else {
				prevnum = kipfwp->fw_number;
				subnumber = 0;
			}
			if (subnumber == 0)
				len = snprintf(one, sizeof(one), "%hu ", kipfwp->fw_number);
			else
				len = snprintf(one, sizeof(one), "%hu.%u ", kipfwp->fw_number, subnumber);
			if (total_len + len > TXT_LEN) {
				syslog(LOG_INFO, "  ** %s", txt);
				total_len = 0;
			}
			strcpy(txt + total_len, one);
			total_len += len;
		}
		if (total_len != 0)
			syslog(LOG_INFO, "  ** %s", txt);
	}
	return 0;
}

/*
 * Remove IP Firewall support: close descriptor for opened socket.
 */
void
kipfw_close(void)
{
	syslog(LOG_INFO, "remove IPv4 Firewall support");
	if (sd != -1) {
		if (close(sd) < 0)
			syslog(LOG_ERR, "kipfw_close: close(FD %d): %m", sd);
		sd = -1;
	}
	free(kipfw);
}
